 /* secure database
  *
  * This file is part of the CRYPTO BONE Project
  * File: openpgp.c
  * Version  : 1.0.3 (All-In-One)
  * License  : BSD
  * Date:      6 May 2016
  *
  * Summary:
  * This program uses cryptlib to encrypt messages to the OpenPGP message format.
  * It also decrypts OpenPGP messages. All encryption and decryption uses a single
  * random password greater than 19 characters and AES.
  * The maximum message size is 50000 bytes.
  *
  * Copyright (c) 2015-2016
  *	Ralf Senderek, Ireland.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
  * 3. All advertising materials mentioning features or use of this software
  *    must display the following acknowledgement:
  *	   This product includes software developed by Ralf Senderek.
  *
  * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  *****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cryptlib.h"
#include "b64.h"

#define MAXINPUT 50000
#define MAXBUF  2*MAXINPUT 
#define KEYLEN 101
#define MAXLINE 200
#define INTERNAL_BUFFER_SIZE MAXINPUT+1024
#define cryptUser CRYPT_UNUSED
#define DEBUG 0

#define OUTFILE "/usr/lib/cryptobone/cryptobone/encryptedmessage.asc"
#define PGP_VERSION "Version: Crypto Bone 1.0 with Cryptlib 3.4.3"

char plaintext[MAXINPUT] ;
char input[MAXBUF];
char output[MAXBUF*4/3];           /* output holds base64 encoded data and exceeds input */
char text[(MAXBUF*4/3)+1024];      /* holds armored data and format info */
char CODE[6];                      /* CRC-24 code (base64-encoded) with = */
char CRC[5];                       /* CRC-24 code (base64-encoded) without = */

char password[KEYLEN]; 
int  passwordLength = 0;
int  status;
int  num;

char FileName[KEYLEN];
FILE *fp;

/***************************************************************************************/
unsigned long crc24(int len) {

      #define CRC24_INIT 0xB704CE
      #define CRC24_POLY 0x1864CFB

      long crc;
      crc = CRC24_INIT;
      int i,j;
      j = 0;
      while (len--) {
           crc ^= (input[j]) << 16;
	   j++;
	   for (i = 0; i < 8; i++) {
	        crc <<= 1;
		if (crc & 0x1000000){
		     crc ^= CRC24_POLY;
		}
	   }
      }	   
      return crc & 0xFFFFFF;
}


/***************************************************************************************/
int encrypt(){
      /*
       * Reads from plaintext and writes the cryptogram to input and base64-encoded to output.
       * output is stored in a file "encryptedmessage.asc"
       */
      
      unsigned int i = 0;
      int j = 0;
      unsigned int bytesCopied = 0;
      unsigned int plaintextLength = strlen(plaintext);
      char BYTES[3];
      unsigned long crcL;

      if (plaintextLength >= MAXINPUT){
           puts("Error: plaintext is too big.");
	   exit(2);
      }

      if (plaintextLength == 0){
           puts("Error: plaintext is empty.");
	   exit(3);
      }

      CRYPT_ENVELOPE cryptEnvelope;
      
      if (DEBUG)  printf("%i bytes to be encrypted: %s \n",plaintextLength, plaintext);

      status = cryptCreateEnvelope( &cryptEnvelope, cryptUser, CRYPT_FORMAT_PGP );
      if (DEBUG)  printf("STATUS: %i \n",status);
      if( status != CRYPT_OK ){
          puts("CryptLib: Envelope cannot be created.");
	  exit(4);
      }
     
      /* set internal buffer size */
      status = cryptSetAttribute( cryptEnvelope, CRYPT_ATTRIBUTE_BUFFERSIZE, INTERNAL_BUFFER_SIZE );
      if (DEBUG)  printf("STATUS (buffsize): %i \n",status);

      /* Add the password */
      passwordLength = strlen(password);
      if (DEBUG)  printf("Using password %s of length %i\n",password,passwordLength);
      status = cryptSetAttributeString( cryptEnvelope, CRYPT_ENVINFO_PASSWORD, password, passwordLength );

      if (DEBUG)  printf("STATUS (pw): %i : (%i) \n",status, passwordLength);
      if (plaintextLength > 1) {
           cryptSetAttribute( cryptEnvelope, CRYPT_ENVINFO_DATASIZE, plaintextLength );
           status = cryptPushData( cryptEnvelope, plaintext, plaintextLength, &bytesCopied );
           if (DEBUG)  printf("STATUS (push): %i \n",status);
           status = cryptFlushData( cryptEnvelope );
           if (DEBUG)  printf("STATUS (flush): %i \n",status);
      
           /* create cryptogram in buffer input */
           status = cryptPopData( cryptEnvelope, input, MAXBUF, &bytesCopied );
           if (DEBUG)  printf("STATUS: %i \n",status);
           if (DEBUG)  printf("encrypted Buffer has %i bytes\n", bytesCopied);

           num = base64encode(output, input, bytesCopied);
           if (DEBUG)  printf("in:\n[%s]\n out:%s\n",input, output);

           /* write result to file */
           fp = fopen(OUTFILE,"w");
           if (!fp){
                printf ("Error: output file is not available");
           }
           else {
                if (DEBUG)  printf ("saving ... %s with %i bytes  \n",output,(int)strlen(output));
                fputs("-----BEGIN PGP MESSAGE-----\n",fp);
		fputs(PGP_VERSION,fp);
		fputs("\n\n",fp);
		/* print armored block */
		for ( i=0; i < strlen(output); i++ ){
		     if ((i>0) && (i%64 == 0)) fputs("\n",fp);
                     fputc(output[i],fp);
		}
		fputs("\n",fp);

		/* calculate CRC code */
		crcL = crc24(bytesCopied);
		for (j=2; j >= 0; j--){
                     BYTES[j] = (crcL%256);  
		     crcL = crcL/256;
		}
                num = base64encode(CRC, BYTES, 3);
		CODE[0] = '=';
		CODE[5] = '\0';
		for (j=1; j < 5; j++){
	             CODE[j] = CRC[j-1];
	        }
		fputs(CODE,fp);
		fputs("\n",fp);
                fputs("-----END PGP MESSAGE-----\n",fp);
                fclose(fp);
           }
      }
      else {
           puts("Error: plain text too short.");
	   exit(3);
      }
      status = cryptDestroyEnvelope( cryptEnvelope );
      if( status != CRYPT_OK ){
          puts("CryptLib: Envelope cannot be destroyed.");
	  exit(5);
      }
      
      return 0;
}

/***************************************************************************************/
int decrypt(){
      /*
       * reads base64 cryptogram from input file (buffer output) decodes to input and 
       * writes the decrypted message to plaintext
       */
      
      int bytesCopied = 0;
      int i;
      char BYTES[3];
      char testchar;
      unsigned long crcL;
      CRYPT_ENVELOPE cryptEnvelope;

      /* decodes output and writes the result to input  */
      num = base64decode(input, output);
      if (DEBUG) printf("base64: %i bytes received in input ",num);
     
      /* check CRC code */
      base64decode(BYTES, CRC);
      if (DEBUG)  printf("CRC: %s\n",CRC);
      /* calculate CRC code */
      crcL = crc24(num);
      for (i=2; i >= 0; i--){
             testchar = (crcL%256);  
	     if (testchar != BYTES[i]) {
                  puts("Decryption Error: CRC is invalid.");
		  exit(1);
	     }
             crcL = crcL/256;
      }
      
      if (strlen(input) > 1){
           /* decrypt */
           status = cryptCreateEnvelope( &cryptEnvelope, cryptUser, CRYPT_FORMAT_AUTO );
	   
           /* set internal buffer size */
           status = cryptSetAttribute( cryptEnvelope, CRYPT_ATTRIBUTE_BUFFERSIZE, INTERNAL_BUFFER_SIZE );
           status = cryptPushData( cryptEnvelope, input, MAXBUF, &bytesCopied );
	   
           if (DEBUG)   printf("%i bytes input to decrypt\n", bytesCopied);

	   passwordLength = strlen(password);
           if (DEBUG)  printf("Using password %s of lenght %i\n",password,passwordLength);
           cryptSetAttributeString( cryptEnvelope, CRYPT_ENVINFO_PASSWORD, password, passwordLength );
           status = cryptFlushData( cryptEnvelope );
           if (DEBUG)  printf("STATUS %i\n",status);
           status = cryptPopData( cryptEnvelope, plaintext, MAXBUF, &bytesCopied );
           if (status != CRYPT_OK){
		if (status == -32){
                     puts("Error: input data is corrupt.");
		}
		if (status == -33){
                     puts("Error: integrity check failed.");
		}
                if (DEBUG)   printf("decryption status is %i \n",status);
	        exit(1);
           }

           if (DEBUG)   printf("decryption status is %i \n",status);
      }

      status = cryptDestroyEnvelope( cryptEnvelope );
      if( status != CRYPT_OK ){
          puts("CryptLib: Envelope cannot be destroyed.");
	  exit(5);
      }
      if (DEBUG)   printf("Decryption result: \n%s \nwith %i bytes \n", plaintext, bytesCopied);

      return 0;
}

/***************************************************************************************/
int check_crypto(){
      /*
       * performs an unsuccessful decryption on an input buffer and reports the ALGO used. 
       */
      char ModeInfo[4];
      int Algo = 0;
      int Mode = 0;
      int Keysize = 0;
      int bytesCopied = 0;
      CRYPT_ENVELOPE cryptEnvelope;

      /* decodes output and writes the result to input  */
      num = base64decode(input, output);
      if (DEBUG) printf("base64: %i bytes received in input ",num);
     
      if (strlen(input) > 1){
           /* decrypt with wrong key */
           status = cryptCreateEnvelope( &cryptEnvelope, cryptUser, CRYPT_FORMAT_AUTO );
           
	   /* set internal buffer size */
           status = cryptSetAttribute( cryptEnvelope, CRYPT_ATTRIBUTE_BUFFERSIZE, INTERNAL_BUFFER_SIZE );
           status = cryptPushData( cryptEnvelope, input, MAXBUF, &bytesCopied );
	   
           if (DEBUG)   printf("%i bytes input to decrypt\n", bytesCopied);

	   /* use wrong key deliberately */
           strlcpy(password,"12345",6);
	   passwordLength = strlen(password);
           if (DEBUG)  printf("Using wrong password %s of lenght %i\n",password,passwordLength);
           cryptSetAttributeString( cryptEnvelope, CRYPT_ENVINFO_PASSWORD, password, passwordLength );
           status = cryptFlushData( cryptEnvelope );
			 
           /* check the encryption information */			 
	   status=cryptGetAttribute(cryptEnvelope, CRYPT_CTXINFO_ALGO, &Algo);
	   status=cryptGetAttribute(cryptEnvelope, CRYPT_CTXINFO_MODE, &Mode);
	   status=cryptGetAttribute(cryptEnvelope, CRYPT_CTXINFO_KEYSIZE, &Keysize);
	   if (Mode == 1) strlcpy(ModeInfo,"ECB",4);
	   if (Mode == 2) strlcpy(ModeInfo,"CBC",4);
	   if (Mode == 3) strlcpy(ModeInfo,"CFB",4);
           Keysize = Keysize*8;
	   if (DEBUG) {  
	        printf("mode: %i (%s)\n",Mode, ModeInfo);
	   }
           if (DEBUG) {  
	        printf("Keysize: %i bit\n",Keysize);
           }		
           if (DEBUG) {
	        printf("ALGO: %i\n",Algo);
	   }	
	   if (Algo == 2) {
                printf("%s\n","3DES encrypted data");
           }
	   if (Algo == 4) {
                printf("%s\n","CAST5 encrypted data");
           }
	   if (Algo == 8) {
                printf("%s\n","AES encrypted data");
           }
      }
      else {
           puts("Error: no input data.");
	   exit(1);
      }

      status = cryptDestroyEnvelope( cryptEnvelope );
      if( status != CRYPT_OK ){
          puts("CryptLib: Envelope cannot be destroyed.");
	  exit(5);
      }
      return 0;
}
/***************************************************************************************/
int analyse_text_input() {

     unsigned int Start = 0;
     unsigned int End = strlen(text);
     char PGP_BEGIN[30] = "-----BEGIN PGP MESSAGE-----\n";
     char PGP_END[30] = "-----END PGP MESSAGE-----\n";
     unsigned int i,j,length;
     int begin = 0;
 
     /* find the armored data in a PGP message (text) and copy to output */
     /* find -----BEGIN */
     length = strlen(text);
     i = 0;
     while ( (begin == 0) && (i<length) ) {
          while ((i < length) && (text[i] != '-')) {
               i++;
          }
          if (i < length) begin = 1;
	  j = 0;
	  while ((j < strlen(PGP_BEGIN) -1 ) && (begin == 1) && (i < length)) {
	       if (text[i] != PGP_BEGIN[j]){
                    begin = 0;
	       }
	       /* if (DEBUG) printf("%i %c %c %i\n",i,text[i],PGP_BEGIN[j],begin);*/
	       i++;
	       j++;
	  }
	  if (begin == 1) {
	       /* skip next line if version line exists */
	       i++;
	       if (text[i] == 'V'){
	            /* read to end of line */
                    while ((i < length) && (text[i] != '\n')){
                         i++;
                    }
	       }
	       i++;
	       Start = i;
	  }
     }
     /* search for -----END */
     begin = 0;
     while ((begin == 0) && (i < length)) {
          while ((i < length) && (text[i] != '-')) {
               i++;
          }
          if (i < length) begin = 1;
	  j = 0;
	  while ((j < strlen(PGP_END) -1 ) && (begin == 1) && (i < length)) {
	       if (text[i] != PGP_END[j]){
                    begin = 0;
	       }
	       /* if (DEBUG) printf("%i %c %c %i\n",i,text[i],PGP_END[j],begin);*/
	       i++;
	       j++;
	  }
	  if (begin == 1) {
	       End = i - strlen(PGP_END) -6;
               /* determine CRC code */
	       i = End + 2;
	       for (j = 0; j < 4; j++) {
	            CRC[j] = text[i];
		    i++;
	       }
	       /* copy part Start to End to output */
	       j = 0;
	       i = Start;
	       while (i < End) {
                    if (text[i] != '\n') {
		         output[j] = text[i];
		         j++;
		    }
		    i++;
	       }
	       output[j] = '\0';
	       return 0;	        
          }
     }	  
     return 1;
}



/***************************************************************************************/
int main (int argc, char *args[]) {

    char *strptr;
    char pattern[5];
    char src[MAXLINE];
    char Line[MAXLINE];
    char NewLine[MAXLINE];
    char *sptr;

    int pos = 0;
    int i = 0;
    int j = 0;
    status = cryptInit();

    if( status != CRYPT_OK ){
          /* cryptlib initialisation failed */;
          puts("CryptLib initialisation failed.");
	  exit(2);
    }
    
    cryptAddRandom( NULL, CRYPT_RANDOM_SLOWPOLL );
    
    if (argc > 3){

          if (strcmp(args[1],"--password") == 0) {
               strlcpy(password,args[2], KEYLEN);
	       if (strlen(password) < 20) {
                    printf("Incorrect message key length : %i (>= 20 required)\n",(int)strlen(password));
		    exit(6);
	       }
	  
	  }
          
          if (strcmp(args[3],"--encrypt") == 0) {
	       /* INPUT is piped into this program, read stdin, and if less than MAXINPUT,
	       *  copy to plaintext and proceed to write the encrypted message to 
	       *  /cbb/cryptobone/encryptedmessage.asc
	       */
               status = read(0,plaintext,MAXINPUT);
	       plaintext[MAXINPUT-1] = '\0';
	       status = encrypt();
               if (status != 0) {
                    puts("Error: encryption failed.");
                    exit(1);
               }
	       exit(0);
          }
          
          
	  if (strcmp(args[3],"--decrypt") == 0){
               if (argc > 4){
                    fp = fopen(args[4],"r");
                    if (!fp){
                         printf ("Error: input file is not readable\n");
			 exit(1);
                    }
                    else {
		         strlcpy(text,"",MAXBUF*4/3);
                         while (!feof(fp)){
                              sptr = fgets(Line,MAXLINE,fp);
                              /* remove \r in Line */
                              i = 0;
                              j = 0;
                              while (i < strlen(Line)){
                                   if (Line[i] != '\r') {
                                        NewLine[j] = Line[i];
                                        j++;
                                   }
                                   i++;
                              }
                              NewLine[j] = '\0';
			      if (!feof(fp))  strlcat(text, NewLine, MAXBUF*4/3);
			 }
                         fclose(fp);
                    }

                    status = analyse_text_input();
                    if (status == 0) {
	                 status = decrypt();
		         if ( status == 0){
			      /* sucessful decryption */
		              /* determine output file name from args[4] or stdout  */
			      strlcpy(pattern,".asc",5);
                              strlcpy(src,args[4],MAXLINE);
			      strptr = strstr(src,pattern);
			      pos = strptr - src;
                              /* printf ("found pos %i\n",pos); */
                              if (strptr != 0) { 
                                   strncpy(FileName, args[4] ,pos);
			           /* printf("Filename: %s",FileName) ; */ 
                                   /* write result to file */
                                   fp = fopen(FileName,"w");
                                   if (!fp){
                                        printf ("Error: output file is not available");
                                   }
                                   else {
                                        if (DEBUG)  printf ("saving ... %s with %i bytes  \n",plaintext, (int) strlen(plaintext));
                                        fputs(plaintext,fp);
                                        fclose(fp);
                                   }
			      }
			      else {
	                           /* use stdout */
			           printf("%s",plaintext);
			      }
                         }
                         else {
	                      puts("Error: decryption failed.");
                              exit(1);
                         }			      
		    }
		    else {
	                 puts("Error: No valid PGP data found.");
			 exit(1);
		    }
	       }
	       else {
	            puts("Error: input file is missing");
		    exit(1);
	       }
          }
    }
    else {
          if (argc == 3){
	       if (strcmp(args[1],"--check") == 0){
               
                    fp = fopen(args[2],"r");
                    if (!fp){
                         printf ("Error: input file is not readable\n");
		         exit(1);
                    }
                    else {
		         strlcpy(text,"",MAXBUF*4/3);
                         while (!feof(fp)){
                              sptr = fgets(Line,MAXLINE,fp);
                              /* remove \r in Line */
                              i = 0;
                              j = 0;
                              while (i < strlen(Line)){
                                   if (Line[i] != '\r') {
                                        NewLine[j] = Line[i];
                                        j++;
                                   }
                                   i++;
                              }
                              NewLine[j] = '\0';
			      if (!feof(fp))  strlcat(text, NewLine, MAXBUF*4/3);
		         }
                         fclose(fp);
                    }

                    status = analyse_text_input();
                    if (status == 0) {
		         status = check_crypto();
		    }
		    else {
	                 puts("Error: No valid PGP data found.");
			 exit(1);
                    }
               exit(0);     			 
               }		    
          }
	  else {
               printf("usage: openpgp --check infile | --password password [ --encrypt |  --decrypt infile ] \n");
	  }
    }


    status = cryptEnd();
    if( status != CRYPT_OK ){
          /* cryptlib shutdown failed */;
          puts("Cryptlib failed to shut down. Some sensible data may remain.\n");
    }
    return 0;
}
